#include <applicat.h>
#include <date.h>
#include <form.h>
#include <modaut.h>
#include <msksheet.h>
#include <prefix.h>
#include <printer.h>
#include <utility.h>
#include <urldefid.h>

#ifndef _DEMO_
#include <hlapi_c.h>
#endif

#include "ba1.h"
#include "ba1500.h"
#include "ba1500a.h"
#include "ba1600.h"

#define USERADR    26952
#define AGAADR     26953
#define PRASSIADR  26954
#define PROCOMADR  26956
#define K1         0x4500
#define LBYTEMASK  0x00FF
#define UBYTEMASK  0xFF00

#define  BITTEST(w,p) (((w) & (0x0001 << (p))) != 0)
#define  BITSET(w,p,v)  ((v) ? ((w) |= (0x0001 << (p))) : ((w) &= (~(0x0001 << (p)))))
           
TDongle::TDongle() 
       : _type(_user_key), _serno(0xFFFF), _dirty(FALSE), 
         _max_users(1), _year_assist(1997)
{
  memset(_eprom, 0, sizeof(_eprom));
}  

void TDongle::garble(word k[4]) const
{   
#ifndef _DEMO_

#if XVT_OS == XVT_OS_SCOUNIX
  INT_OFF();
  HL_ON(_port, ModAd);
  K_EYE(_port, k, HLBLOCK);
  HL_OFF(_port);
  INT_ON();
#else                 
  HL_CODE(EYECAST k, 1); // HLBLOCK = 1
#endif

#endif  // _DEMO_
}

bool TDongle::already_programmed() const
{     
#ifndef _DEMO_
  word data[4];
  memcpy(data, &_eprom[60], sizeof(data));
  garble(data);
  
  if (data[0] < 1997 || data[0] > 2997)
    return FALSE;
    
  if (data[1] == 0 || data[1] >= 10000)  
    return FALSE;
  
  const TDate today(TODAY);
  const long& giulio = (const long&)data[2];
  const long yyyymmdd = today.julian2date(giulio);
  const TDate d(yyyymmdd);
  if (d.year() < 1997 || d > today)
    return FALSE;
#endif // _DEMO_    
 return TRUE;  
}
           
bool TDongle::login(bool test_all_keys)
{
  bool ok = TRUE;
  _type = _user_key;

#ifdef _DEMO_
  _serno = 0;
#else
#if XVT_OS == XVT_OS_SCOUNIX
  if (test_all_keys)
  {
    if (Hl_Port(AGAADR) != 0)
      _type = _aga_key;
    else
      if (Hl_Port(PRASSIADR) != 0)
        _type = _prassi_key;
      else
        if (Hl_Port(PROCOMADR) != 0)
          _type = _prcom_key;
  }        
  _port = Hl_Port(USERADR);
#else
  if (test_all_keys)
  {
    HL_LOGOUT();
    if (HL_LOGIN(AGAADR, DONT_CARE, REFKEY, VERKEY) == STATUS_OK) 
      _type = _aga_key;
    else
    {
      HL_LOGOUT();
      if (HL_LOGIN(PRASSIADR, DONT_CARE, REFKEY, VERKEY) == STATUS_OK)
        _type = _prassi_key;
      else
      {
        HL_LOGOUT();
        if (HL_LOGIN(PROCOMADR, DONT_CARE, REFKEY, VERKEY) == STATUS_OK)
          _type = _procom_key;
      }
    }
  }  
  HL_LOGOUT();
  ok = HL_LOGIN(USERADR, DONT_CARE, REFKEY, VERKEY) == STATUS_OK;
  
  if (ok)
  {
    HL_READBL((char*)_eprom);

    word data[4];
    memcpy(data, _eprom, sizeof(data));
    garble(data);
    
    if (data[0] == 0xFAE8)
      _serno = data[1];
    else
    {                  
      if (data[0] == 0x3283) // chiave programmatori !!
      {         
#ifdef DBG    
//        _type = _aga_key; // eliminato perch giudicato troppo pericoloso
                            // se la DBG version arriva nelle mani di qualcuno
                            // ci inchiappetta tutti i programmi senza darci una lira!
                            
#endif // DBG       
        _serno = 0;  
      }  
    }  
  }  
#endif // SCOUNIX
#endif // _DEMO_

  if (ok) 
  {
    if (_serno > 0)
    {  
      const bool already = already_programmed();
      
      _module.reset();    
      _module.set(0, TRUE);
      const int last_word = already ? 12 : 4;                    
      word data[4];

      // Legge flag di attivazione dei moduli
      for (int i = 0; i < last_word; i += 4)
      {
        memcpy(data, &_eprom[48+i], sizeof(data));
        garble(data);
        if (data[3] == _serno) // Validate block
        {
          for (int j = 0; j < 3; j++)
          {
            word parola = data[j] ^ _serno;
            if (parola)
            {
              for (int b = 15; b >= 0; b--)
              {
                if (BITTEST(parola, b))             
                {
                  const word bit = i * 12 + j * 16 + b;
                  _module.set(bit + 1);
                }  
              }  
            }  
          }
        }
        else
          break;
      }
      
      // Legge anno di assitenza e numero massimo di utenti
      memcpy(data, &_eprom[60], sizeof(data));
      garble(data);      
      
      if (already)
      {
        _year_assist = data[0];
        _max_users = data[1];
        const long& giulio = (const long&)data[2];
        const long yyyymmdd = _last_update.julian2date(giulio);
        _last_update = yyyymmdd;
      }
      else
      {
        _max_users = 1;
        _last_update = TDate(TODAY);
        _year_assist = _last_update.year();
        _dirty = TRUE;
      }
    }  
    else
    {                   
      if (_type != _user_key)                
      {
        _module.set(MAX_AUT); 
        _module.set();    // Activate all modules
      }  
      _max_users = 1;
      _last_update = TDate(TODAY);
      _year_assist = _last_update.year();
    }  
  }  

  return ok;  
}

bool TDongle::logout()
{
#ifndef _DEMO_
  HL_LOGOUT();
#endif  
  _serno = 0xFFFF;
  return TRUE;
}

bool TDongle::write_octect(word reg, word k[4]) const
{                     
#ifdef _DEMO_
  return FALSE;
#else
  int err = STATUS_OK;
  for (word r = 0; r < 4; r++)
  {           
    const word address = reg+r;
    err = HL_WRITE(address, k[r]);
    if (err != STATUS_OK)                                     
    {
      error_box("Errore di scrittura sul registro %u", address);
      break;
    }  
#ifdef DBG
    word test;
    HL_READ(address, (int*)&test);
    if (test != k[r])
      error_box("Errore di scrittura sul registro %u", address);
#endif // DBG     
  }  
  return err == STATUS_OK;
#endif // _DEMO_
}

bool TDongle::burn()
{
#ifdef _DEMO_
  return FALSE;
#endif // _DEMO_
  
  if (!dirty() || _serno == 0 || _serno == 0xFFFF)
    return FALSE;
    
  word data[4];

  const TDate today(TODAY);   
  const bool already = already_programmed();
  if (already)  
  {       
    memcpy(data, &_eprom[60], sizeof(data));
    garble(data);                         
    if (data[0] < 1997 || data[0] > 2997)
      return error_box("On Line Assistance error.");
    if (data[1] == 0 || data[1] >= 10000)
      return error_box("Bad users number.");
    const long& val = (const long&)data[2];
    const long yyyymmdd = today.julian2date(val);
    const TDate date(yyyymmdd);
    if (date > today)
      return error_box("Too late sir: key has already expired!");
  }
  
  data[0] = _year_assist;
  data[1] = _max_users;
  long& val = (long&)data[2];
  val = today.date2julian();
  garble(data);
  write_octect(60, data);
  _last_update = today;
  
  word module = 1;
  for (int octect = 0; octect < 3; octect++)  
  {                                       
    for(int parola = 0; parola < 3; parola++)  
    {        
      word& p = data[parola];
      p = 0;
      for (int bit = 0; bit < 16; bit++)
      {          
        if (active(module))
          BITSET(p, bit, TRUE);
        module++;
      }
      p ^= _serno;
    }
    data[3] = _serno;
    garble(data);
    write_octect(48 + 4*octect, data);
  }
  
  _dirty = FALSE; 
  return TRUE;
}                   

///////////////////////////////////////////////////////////

TInformazione_moduli::TInformazione_moduli()
{
  int mod = 0;
  TString s(256);
  TToken_string t;
  TString_array descs;
  
  _unassigned_modules = 0;
  ifstream  in("prassi.aut");
  while (!in.eof() && in.good())
  {
    in.getline(s.get_buffer(), s.size());
    s.trim();
    if (s.empty()) 
      break;
    const int l = s.len();
    if (l > 2)
    {
//      t = format("%-40s", (const char*)s.right(l-3));
      t = s.mid(3);
      t.add(s.left(2));
    }
    else
    {
      t = " "; t.add(" ");
      _unassigned_modules++;
    }
    t.add(mod++); 
    s = t.get(0);
    descs.add(s);
    s.upper();
    t.add(s, 0);
    _infos.add(t);
  }
  
  if (in.bad() && !in.eof())
    error_box("Si e' verificato un errore leggendo il file di descrizione moduli.");
  
  _infos.sort();
  for (int i=0;i<mod;i++)
  {
    TToken_string& riga = _infos.row(i);
    const int mod = riga.get_int(2);
    _index[mod] = i;         
    s = descs.row(mod).get(0);
    riga.add(s, 0);
  }
}

const char* TInformazione_moduli::get_description_by_order(int index)
{ 
  if (index >= 0 && index < _infos.items())
    return _infos.row(index).get(0);
  else
    return "";
}

const char* TInformazione_moduli::get_name_by_order(int index)
{ 
  if (index >= 0 && index < _infos.items())
    return _infos.row(index).get(1);
  else 
    return "";
}

int TInformazione_moduli::get_module_by_order(int index)
{
  if (index >= 0 && index < _infos.items())
    return _infos.row(index).get_int(2);
  else
    return -1;
}

TDongle::KeyType TInformazione_moduli::get_owner(int module) const
{                        
  if (module >= 40 && module <= 46)
    return TDongle::_procom_key;        
    
  const int aut[] = { CMAUT, ATAUT, POAUT, AFAUT, TCAUT, TMAUT, 
                      VEAUT, MGAUT, ORAUT, EFAUT, DBAUT, SVAUT, -1 };
  for (int a = 0; aut[a] >= 0; a++)
    if (module == aut[a])
      return TDongle::_aga_key;
      
  return TDongle::_prassi_key;
}

bool TInformazione_moduli::test_owner(int module, TDongle::KeyType owner) const
{
  TDongle::KeyType ow = get_owner(module);
  bool yes = owner == ow || (owner == TDongle::_aga_key && ow == TDongle::_prassi_key);
  return yes;
}

const char* TInformazione_moduli::get_description(int module)
{ return get_description_by_order(_index[module]); }

const char* TInformazione_moduli::get_name(int module)
{ return get_name_by_order(_index[module]); }

int TInformazione_moduli::get_index(int module)
{ return _index[module]; }


///////////////////////////////////////////////////////////
// TForm richiesta attivazione
///////////////////////////////////////////////////////////

class TRequest_form : public TForm
{ 
public:             
  void update_serno(const TString& serno);
  void update_dongle(const TDongle& dongle);
  void print_footer();

  TRequest_form(const char* name);
  virtual ~TRequest_form() { }
}; 

TRequest_form::TRequest_form(const char* name)
             : TForm(name)
{                                   
  const TPrint_section& header = section('H', odd_page);
  const TPrint_section& footer = section('F', odd_page);
  
  int riga[66];
  int y1 = header.height()-2;
  int y2 = printer().formlen() - footer.height();
  if ((y2-y1) & 0x1) y2--; // Forza un numero pari di righe
  
  for (int i = 0; ; i++)
  {
    riga[i] = y1 + 2*(i+1);
    if (riga[i] >= y2-1)
    {
      riga[i] = 0;
      break;
    }
  }      
  genera_intestazioni(odd_page, y1+1);
  genera_fincatura(odd_page, y1, y2, riga);
} 

void TRequest_form::update_serno(const TString& num)
{
  const TDate today(TODAY);
  TForm_item& data = find_field('H', odd_page, 101);
  data.set(today.string());
  TForm_item& serno = find_field('H', odd_page, 102);
  serno.set(num);
}

void TRequest_form::update_dongle(const TDongle& dongle)
{
  TForm_item& year = find_field('H', odd_page, 103);
  TString str; 
  str << "31-12-" << (int)dongle.year_assist();
  year.set(str);
  
  TForm_item& user = find_field('H', odd_page, 104);
  str.format("%d", dongle.max_users());
  user.set(str);
}

void TRequest_form::print_footer()
{
  set_footer(odd_page, TRUE);
  printer().formfeed();
}

///////////////////////////////////////////////////////////
// TAttivazione_moduli
///////////////////////////////////////////////////////////

class TAttivazione_moduli : public TApplication
{
  TMask* _msk;
  TInformazione_moduli* _im;

  TDongle _dongle;
  word _serno;
  
  word _wanted_users, _wanted_assist;

protected:
  virtual bool create() ;                         
  virtual bool destroy() ;                        
  virtual bool use_files() const { return FALSE; }
  virtual bool menu(MENU_TAG);
  
  void garble(word k[4]) const;
  void garble(word n, TString& str) const;
  void garble_user(word u, TString& str) const;
  void garble_year(word u, TString& str) const;

  word& serno() { return _serno; }
  
  void generate_key();
  int build_sheet(bool on = TRUE);
  void build_key_column();
  
  bool burn_dongle();
  TDongle& dongle() { return _dongle; }
  
  static void keyext(const TString & s, word * v);
  static void encode_second_key();
  
  static bool user_hnd(TMask_field& f, KEY k);
  static bool date_hnd(TMask_field& f, KEY k);
  static bool serno_hnd(TMask_field& f, KEY k);
  static bool decode_hnd(TMask_field& f, KEY k);
  static bool activate_hnd(TMask_field& f, KEY k);
  
  static bool k_notify(TSheet_field& f, int r, KEY k);
  static bool change_users_handler(TMask_field&, KEY key);
  static bool maxuser_handler(TMask_field& fld, KEY key);
  static bool change_assist_handler(TMask_field&, KEY key);
  static bool assist_handler(TMask_field& fld, KEY key);
  static bool print_handler(TMask_field& fld, KEY key);
  static bool aggiorna_listino_handler(TMask_field& fld, KEY key);
  static bool aga_handler(TMask_field& fld, KEY key);
  
protected:
  void print_request();  
  void print_answer();

public:                       
  virtual void print();

  TAttivazione_moduli() : _msk(NULL), _wanted_users(0), _wanted_assist(0) { }
  virtual ~TAttivazione_moduli() { }
};


HIDDEN TAttivazione_moduli& app() { return (TAttivazione_moduli &)main_app(); }

int TAttivazione_moduli::build_sheet(bool on)
{
  int nmod = 0;
  TSheet_field& sf = (TSheet_field&)_msk->field(F_MODULI);
  for (int i = 0; i < MAX_AUT; i++)
  {
    const TFixed_string d(_im->get_description_by_order(i));
    if (d.blank()) 
      continue;
    TToken_string& riga = sf.row(i);
    riga = d; riga.trim();
    const int module = _im->get_module_by_order(i);
    
    const bool ics = module == 0 || (on && _dongle.active(module));
    if (ics)
    {
      riga.add("X");
      if (module != 0) 
        nmod++;
    } 
    else 
      riga.add(" ");
    riga.add(module);
  }
  return nmod;
}

void TAttivazione_moduli::garble(word k[4]) const
{
  _dongle.garble(k);
}

void TAttivazione_moduli::garble(word n, TString& str) const
{
  const TDate today(_msk->get(F_DT));
  const long val = today.date2julian();
  
  word data[4];
  data[0] = word(_msk->get_int(F_SN));
  data[1] = n;
  data[2] = word(val >> 16);
  data[3] = word(val & 0xFFFF);
  garble(data);
  str.format("%04X%04X", data[0], data[1]);
}

void TAttivazione_moduli::garble_user(word u, TString& str) const
{
  const TDate today(_msk->get(F_DT));
  const long val = today.date2julian();
  
  word data[4];
  data[0] = word(_msk->get_int(F_SN));
  data[1] = u;
  data[2] = word(val >> 16);
  data[3] = word(val & 0xFFFF);
  garble(data);
  str.format("%04X%04X", data[0], data[1]);
}

void TAttivazione_moduli::garble_year(word y, TString& str) const
{
  const TDate today(_msk->get(F_DT));
  const long val = today.date2julian();
  
  word data[4];
  data[0] = y;
  data[1] = word(_msk->get_int(F_SN));
  data[2] = word(val >> 16);
  data[3] = word(val & 0xFFFF);
  garble(data);
  str.format("%04X%04X", data[0], data[1]);
}

void TAttivazione_moduli::build_key_column()
{       
  TSheet_field& sf = _msk->sfield(F_MODULI);
  sf.enable_column(F_KEY, FALSE);

  TString16 tmp;
  FOR_EACH_SHEET_ROW_BACK(sf, i, riga)
  {   
    if (!sf.cell_disabled(i, 1))
    {
      int module = riga->get_int(2);
      garble(module, tmp);
      riga->add(tmp, 3);
    }  
  }
  sf.force_update();
  
  _msk->set(F_USERS, _dongle.max_users());
  _msk->set(F_ASSIST, _dongle.year_assist());
}

bool TAttivazione_moduli::burn_dongle()
{   
  bool ok = _dongle.type() == TDongle::_user_key;
  
  if (ok)
  {
    TString16 str, key;
  
    TSheet_field& sf = _msk->sfield(F_MODULI);
    FOR_EACH_SHEET_ROW(sf, i, riga)
    {
      key = riga->get(3); key.trim();
      if (key.not_empty())
      {
        const int module = riga->get_int(2);
        garble(module, str);
        if (key == str)
        {
          if (!_dongle.active(module) &&
              yesno_box("Confermare l'attivazione del modulo %d:\n%s", 
                        module, riga->get(0)))
          {         
            _dongle.activate(module);
          }              
        }
        else
        {
          if (_dongle.active(module) &&
              yesno_box("Confermare la disattivazione del modulo %d:\n%s", 
                        module, riga->get(0)))
          {              
            _dongle.deactivate(module);
          }  
        }
      }
    }  
    
    if (_dongle.dirty())
    {
      ok = _dongle.burn();
      if (!ok)
        error_box("Impossibile riprogrammare la chiave");
    }
  }
  
  return ok;  
}

bool TAttivazione_moduli::maxuser_handler(TMask_field& fld, KEY key)
{         
  bool ok = TRUE;
  if (key == K_TAB && fld.focusdirty())
  {
    const int utenti = atoi(fld.get());
    if (utenti > 0 && utenti < 10000)
    {                             
      TString16 str;
      app().garble_user(utenti, str);
      fld.mask().set(102, str);
    }
    else
      ok = fld.error_box("Numero di utenti errato");  
  }  
  return ok;
}

bool TAttivazione_moduli::change_users_handler(TMask_field& fld, KEY key)
{             
  if (key == K_SPACE)
  {                                   
    TMask m("Utenti", 1, 40, 5);
    m.add_number(101, 0, "Utenti ", 1, 1, 4, "U");
    m.add_string(102, 0, "Codice ", 1, 2, 8, "U");
    m.add_button(DLG_OK, 0, "", -12, -1, 10, 2);
    m.add_button(DLG_CANCEL, 0, "", -22, -1, 10, 2);
    
    TDongle& dongle = app().dongle();
    if (dongle.type() == TDongle::_user_key)
      m.set(101, dongle.max_users());  
    else
    {
      m.set_handler(101, maxuser_handler);
      m.disable(102);
    }  
    
    if (m.run() == K_ENTER)
    {
      const int utenti = m.get_int(101);
      if (dongle.type() == TDongle::_user_key)
      {
        if (utenti > 0 && utenti < 10000)
        {                  
          if (m.field(102).empty())
          {
            app()._wanted_users = utenti; 
            if (utenti > 0)
              fld.message_box("Premere il bottone '%s'\n"
                              "per richiedere l'attivazione di %d utenti.",
                              fld.mask().field(F_PRINT).prompt(), utenti);
          }
          else
          {
            TString16 str;
            app().garble_user(utenti, str);
            if (str == m.get(102))
            { 
              dongle.set_max_users(utenti);
              fld.mask().set(F_USERS, utenti);
            }  
            else
              fld.error_box("Codice errato");
          }    
        }  
        else
          fld.error_box("Numero di utenti errato");
      }                                           
      else
        app()._wanted_users = utenti;
    }
  }
  return TRUE;
}

bool TAttivazione_moduli::assist_handler(TMask_field& fld, KEY key)
{         
  bool ok = TRUE;
  if (key == K_TAB && fld.focusdirty())
  {
    const int year = atoi(fld.get());
    if (year > 1997 && year <= 2997)
    {                             
      TString16 str;
      app().garble_year(year, str);
      fld.mask().set(102, str);
    }
    else
      ok = fld.error_box("Anno errato");  
  }  
  return ok;
}

bool TAttivazione_moduli::change_assist_handler(TMask_field& fld, KEY key)
{             
  if (key == K_SPACE)
  {                                   
    TMask m("Assistenza", 1, 40, 5);
    m.add_number(101, 0, "Anno   ", 1, 1, 4, "U");
    m.add_string(102, 0, "Codice ", 1, 2, 8, "U");
    m.add_button(DLG_OK, 0, "", -12, -1, 10, 2);
    m.add_button(DLG_CANCEL, 0, "", -22, -1, 10, 2);
    
    TDongle& dongle = app().dongle();
    if (dongle.type() == TDongle::_user_key)
      m.set(101, dongle.year_assist());  
    else
    {
      m.set_handler(101, assist_handler);
      m.disable(102);
    }  
    
    if (m.run() == K_ENTER)
    {
      const word year = m.get_int(101);
      if (dongle.type() == TDongle::_user_key)
      {
        if (year > dongle.year_assist() && year <= 2997)
        {
          if (m.field(102).empty())
          {
            app()._wanted_assist = year;
            fld.message_box("Premere il bottone '%s'\n"
                            "per rinnovare il contratto di assistenza per l'anno %u.",
                            fld.mask().field(F_PRINT).prompt(), year);
          }
          else
          {
            TString16 str;
            app().garble_year(year, str);
            if (str == m.get(102))
            { 
              dongle.set_year_assist(year);
              fld.mask().set(F_ASSIST, year);
            }  
            else
              fld.error_box("Codice errato");
          }    
        }  
        else
          fld.error_box("Anno errato");
      }                                           
      else
        app()._wanted_assist = year;
    }
  }
  return TRUE;
}

bool TAttivazione_moduli::print_handler(TMask_field& fld, KEY key)
{
  if (key == K_SPACE)
    app().print();
  return TRUE;
}

bool TAttivazione_moduli::aga_handler(TMask_field& fld, KEY key)
{
  if (key == K_SPACE)        
  {
    TMask mask("Produttore", 1, 54, 5);
    mask.add_string(DLG_USER, 0, "", 1, 1, 50);
    mask.add_button(DLG_OK, 0, "", -12, -1, 10, 2);
    mask.add_button(DLG_CANCEL, 0, "", -22, -1, 10, 2);
    
    TConfig ini("install.ini", "Main");
    mask.set(DLG_USER, ::decode(ini.get("Producer")));
    if (mask.run() == K_ENTER)
      ini.set("Producer", ::encode(mask.get(DLG_USER)));
  }  
  return TRUE;
}


void TAttivazione_moduli::generate_key()
{
  word ud[4];

#ifndef _DEMO_  
#if XVT_OS == XVT_OS_SCOUNIX
  INT_OFF();
  HL_ON(_port, USERADR);
  ud[0] = HL_RD(_port, 48);
  ud[1] = HL_RD(_port, 49);
  ud[2] = HL_RD(_port, 50);
  ud[3] = HL_RD(_port, 51);
  HL_OFF(_port);
  INT_ON();
#else
  HL_READ(48, (int*)&ud[0]);
  HL_READ(49, (int*)&ud[1]);
  HL_READ(50, (int*)&ud[2]);
  HL_READ(51, (int*)&ud[3]);
#endif
#endif // _DEMO_  
  
  _msk->set(F_K2, format("%04X%04X%04X%04X", ud[0], ud[1], ud[2], ud[3]));
  int nmod  = build_sheet();
  ud[0] = _dongle.number();
  ud[1] = K1 | (nmod & LBYTEMASK);
  
  long & l = (long &) ud[2];
  TDate d(TODAY);
  l = d.year()*10000L + d.month()*100L + d.day();

  garble(ud);
  _msk->set(F_K1, format("%04X%04X%04X%04X", ud[0], ud[1], ud[2], ud[3]));
  _msk->set(F_SN, _dongle.number());
  _msk->set(F_DT, d);      
  
  _msk->set(F_USERS, _dongle.max_users());
  _msk->set(F_ASSIST, _dongle.year_assist());
}

int hexdigit(char c)
{
  return c >= 'A' ? (c - 'A' + 10) : (c - '0');
}

void TAttivazione_moduli::keyext(const TString& s, word * val)
{
  for (int i = 0 ; i < 4; i++)
  {
    const int off = i * 4;
    word & l = val[i];
    
    l = 0;
    for (int j = 0; j < 4; j++)
      l = l * 16 + hexdigit(s[off + j]);
  }
}

bool TAttivazione_moduli::user_hnd(TMask_field & f, KEY k)
{
  const TString16 k4(f.get());
  if (!f.to_check(k) || k4.empty()) 
    return TRUE;
  TMask & m = f.mask();
  word ud1[4], ud2[4];
  const TString16 k3(m.get(F_K3));
  
  keyext(k3, ud1);
  keyext(k4, ud2);
  
  app().garble(ud1);
  const long & l = (long &) ud1[0];
  TDate d(l), d1(TODAY);

  d.addmonth(3);
  if (d < d1)
    return f.error_box("data non valida");
  if ((ud1[2] & UBYTEMASK) != K1 || ud1[3] != app().serno()) 
    return f.error_box("primo codice errato");
    
  app().garble(ud2);
  for (int i = 0; i < 4; i++) ud2[i] ^= app().serno();
  if (ud2[3] != 0) 
    return f.error_box("secondo codice errato");
  TSheet_field& sf = (TSheet_field&) m.field(F_MODULI);
  const int un = app()._im->unassigned();
  for (i = un; i < MAX_AUT; i++)
  {
    const int af = app()._im->get_module_by_order(i) - 1;
    if (af < 0) continue;
    TToken_string& tt = sf.row(i-un);
    tt.add(BITTEST(ud2[af / 16], af % 16) ? "X" : " ", 1);
  }
  if (k == K_ENTER)
  {
    keyext(k4, ud2);
#ifndef _DEMO_    
#if XVT_OS == XVT_OS_SCOUNIX
    word port = app()._dongle.port();
    INT_OFF();
    HL_ON(port, ModAd);
    HL_WR(port, 48, ud2[0]);
    HL_WR(port, 49, ud2[1]);
    HL_WR(port, 50, ud2[2]);
    HL_WR(port, 51, ud2[3]);
    HL_OFF(port);
    INT_ON();
#else
    HL_WRITE(48, ud2[0]);
    HL_WRITE(49, ud2[1]);
    HL_WRITE(50, ud2[2]);
    HL_WRITE(51, ud2[3]);
#endif
#endif // _DEMO_
  } 
  else
    sf.force_update();
  return TRUE;
}

bool TAttivazione_moduli::serno_hnd(TMask_field& f, KEY k)
{
  if (k == K_TAB && f.focusdirty())
  {
    app().build_key_column();
  }
  return TRUE;
}

bool TAttivazione_moduli::date_hnd(TMask_field& f, KEY k)
{    
  if (k == K_TAB && f.focusdirty())
  {       
    const TDate data(f.get());
    const TDate oggi(TODAY);
    if (data > oggi)
      return f.error_box("La data deve essere antecedente a quella di sistema");
      
    const TDate& last_update = app()._dongle.last_update();
    if (data < last_update)
      return f.error_box("La data deve essere successiva al %s", last_update.string());
    
    const int year_assist = app()._dongle.year_assist()+1;
    if (data.year() > year_assist)  
      return f.error_box("La data deve essere antecedente al 31-12-%d", year_assist);
  }
  return TRUE;
}

void TAttivazione_moduli::encode_second_key()
{
  TMask * m = app()._msk;
  word    ud1[4], ud2[4];
  int     nmod = 0;
  
  for (int i = 0; i < 4; i++) ud2[i] = 0;
  TSheet_field& sf = (TSheet_field&) m->field(F_MODULI);
  const int un = app()._im->unassigned();
  for (i = un; i < MAX_AUT; i++)
  {
    const int af = app()._im->get_module_by_order(i) -1;
    if (af < 0) continue;
    TToken_string& tt = sf.row(i-un);
    if (tt.get(1)[0] == 'X')
    {
      BITSET(ud2[af / 16], af % 16, TRUE);
      nmod++;
    }
  }
  for (i = 0; i < 4; i++) ud2[i] ^= app().serno();

  const TDate d(m->get(F_DT));
  long & l = (long &) ud1[0];
  
  l = d.year()*10000L + d.month()*100L + d.day();
  ud1[2] = K1 | (nmod & UBYTEMASK);
  ud1[3] = app().serno();
  
  app().garble(ud1);
  m->set(F_K3, format("%04X%04X%04X%04X", ud1[0], ud1[1], ud1[2], ud1[3]));

  app().garble(ud2);
  m->set(F_K4, format("%04X%04X%04X%04X", ud2[0], ud2[1], ud2[2], ud2[3]));
}

bool TAttivazione_moduli::decode_hnd(TMask_field & f, KEY k)
{
  const TString16 k2(f.get());
  if (!f.to_check(k) || k2.empty()) 
    return TRUE;
  
  TMask& m = f.mask();
  word ud1[4], ud2[4];
  const TString16 k1(m.get(F_K1));
  
  keyext(k1, ud1);
  keyext(k2, ud2);

  app().garble(ud1);
  long & l = (long &) ud1[2];
  const TDate  d(l);
  app().serno() = ud1[0];
  if ((ud1[1] & UBYTEMASK) != K1)
    return f.error_box("primo codice errato");
  
  app().garble(ud2);
  for (int i = 0; i < 4; i++) ud2[i] ^= app().serno();
  if (ud2[3] != 0)
    return f.error_box("secondo codice errato");
  m.set(F_SN, app().serno());
  m.set(F_DT, d);
  m.set(F_K3,k1);
  m.set(F_K4,k2);
  TSheet_field& sf = (TSheet_field&) m.field(F_MODULI);
  const int un = app()._im->unassigned();
  for (i = un; i < MAX_AUT; i++)
  {
    const int af = app()._im->get_module_by_order(i) -1;
    if (af < 0) continue;
    TToken_string& tt = sf.row(i-un);
    tt.add(BITTEST(ud2[af / 16], af % 16) ? "X" : " ", 1);
  }
  encode_second_key();
  sf.force_update();
  return TRUE;
}

bool TAttivazione_moduli::activate_hnd(TMask_field & f, KEY k)
{
  encode_second_key();
  return TRUE;
}

bool TAttivazione_moduli::k_notify(TSheet_field& f, int r, KEY k)
{
  if (k == K_INS || k == K_DEL)
    return FALSE;
  return TRUE;
}

bool TAttivazione_moduli::aggiorna_listino_handler(TMask_field& fld, KEY key)
{       
  bool ok = TRUE;
  if (key == K_SPACE)
  {
    TInstall_ini ini;    
    TFilename name = ini.get("DiskPath");
    name.add(ini.default_name());
    while (!fexist(name))
    {
      TMask msk("Aggiornamento listino", 1, -1, -1, 60, 4);
      msk.add_string(DLG_USER, 0, "Percorso", 1, 1, 50);
      msk.add_button(DLG_CANCEL, 0, "", -12, -1, 10, 2);
      msk.add_button(DLG_OK, 0, "", -22, -1, 10, 2);
      msk.set(DLG_USER, name);
      if (msk.run() == K_ENTER)
        name = msk.get(DLG_USER);
      else
      {
        ok = FALSE;
        break;
      }  
    }
    if (ok)
      ok = ini.update_prices(name);
  }  
  return ok;
}

void TAttivazione_moduli::print_request()
{
  TRequest_form form("ba1500a");
  TForm_item& num    = form.find_field('B', odd_page, 101);
  TForm_item& descr  = form.find_field('B', odd_page, 102);
  TForm_item& codice = form.find_field('B', odd_page, 103);
  TForm_item& price  = form.find_field('B', odd_page, 104);
  TForm_item& manut  = form.find_field('B', odd_page, 105);
      
  TInstall_ini ini;    
  const bool special = _wanted_assist > _dongle.year_assist() || _wanted_users != 0;
  const word users = max(_wanted_users, _dongle.max_users());

  TSheet_field& sheet = _msk->sfield(F_MODULI);
  for (TDongle::KeyType owner = TDongle::_aga_key; 
       owner <= TDongle::_procom_key; 
       owner = TDongle::KeyType(owner+1))
  {
    int found = 0;
    TString str;
    
    FOR_EACH_SHEET_ROW(sheet, r, row)
    { 
      const int module = _im->get_module_by_order(r);
      const bool active = _dongle.active(module);
      const bool wanted = *row->get(1) > ' ';

      if (!special && active)
        continue;
      if (!wanted && !active)  
        continue;
      if (_im->get_owner(module) != owner)
        continue;
      
      if (found == 0)
      {
        printer().open();      
        form.update_serno(_msk->get(F_SN));
        form.update_dongle(_dongle);
      }
        
      str.format("%3d", ++found);
      num.set(str);
        
      str = row->get(0);
      descr.set(str);
        
      str = active ? "*ATTIVO*" : "";
      codice.set(str);
      
      str = _im->get_name(module);
      real full, assist;    
      ini.prices(str, users, full, assist);
      if (active)
      {
        real old_full, old_assist;
        ini.prices(str, _dongle.max_users(), old_full, old_assist);
        full -= old_full;
        assist -= old_assist;
      }
      else
        assist = 0.0;
      
      price.set(full.string());
      manut.set(assist.string());

      form.print(-1);
    }   
    
    if (special && found > 0)
    {    
      codice.set("");
      price.set("");
      manut.set("");
      if (_wanted_users != 0)
      {
        str.format("%3d", ++found);
        num.set(str);
         
        str.format("Variazione del numero utenti da %u a %u",
                   _dongle.max_users(), _wanted_users);
        descr.set(str);
        form.print(-1);
      }
      if (_wanted_assist > _dongle.year_assist())
      {
        str.format("%3d", ++found);
        num.set(str);
        str.format("Contratto di assistenza per l'anno %u", _wanted_assist);
        descr.set(str);
        form.print(-1);
      }
    }
    
    if (found)
    {
      form.print_footer();
      printer().close();
    }  
  }
}

void TAttivazione_moduli::print_answer()
{               
  TString str;
  int found = 0;
  
  TRequest_form form("ba1500a");
  TForm_item& num = form.find_field('B', odd_page, 101);
  TForm_item& descr = form.find_field('B', odd_page, 102);
  TForm_item& codice = form.find_field('B', odd_page, 103);

  TSheet_field& sheet = _msk->sfield(F_MODULI);
  FOR_EACH_SHEET_ROW(sheet, r, row) if (*row->get(1) > ' ')
  { 
    const int module = _im->get_module_by_order(r);
    if (module != BAAUT)
    {
      if (found == 0)
      {
        printer().open();      
        form.update_serno(_msk->get(F_SN));
      }

      str.format("%3d", ++found);
      num.set(str);
        
      str = row->get(0);
      descr.set(str);
      
      str = row->get(3);
      codice.set(str);

      form.print(-1);
    }
  }             
  
  if (_wanted_users != 0 || _wanted_assist != 0)
  {
    if (found == 0)
    {
      printer().open();      
      form.update_serno(_msk->get(F_SN));
    }
    if (_wanted_users != 0)
    {
      str.format("%3d", ++found);
      num.set(str);
      str.format("Variazione del numero utenti da %u a %u",
                 _dongle.max_users(), _wanted_users);
      descr.set(str);
      garble_user(_wanted_users, str);
      codice.set(str);
      form.print(-1);
    }
    if (_wanted_assist != 0)
    {
      str.format("%3d", ++found);
      num.set(str);
      str.format("Contratto di assistenza per l'anno %u", _wanted_assist);
      descr.set(str);
      garble_year(_wanted_assist, str);
      codice.set(str);
      form.print(-1);
    }
  }

  if (found)
  {
    form.print_footer();
    printer().close();
  }  
}


void TAttivazione_moduli::print()
{
  if (_dongle.type() == TDongle::_user_key)
    print_request();
  else              
    print_answer();
}

bool TAttivazione_moduli::create()
{
  bool ok = _dongle.login(TRUE);
  if (!ok)
    return FALSE;

  _im = new TInformazione_moduli;
  
  disable_menu_item(M_FILE_NEW);
  disable_menu_item(M_FILE_REVERT);
  disable_menu_item(M_FILE_PG_SETUP);

  _msk = new TMask("ba1500a") ;
  
  _msk->set_handler(F_CHANGEUSERS, change_users_handler);
  _msk->set_handler(F_CHANGEASSIST, change_assist_handler);
  _msk->set_handler(F_PRINT, print_handler);
  if (_dongle.type() == TDongle::_aga_key)
  {                                  
    _msk->show(F_AGA);
    _msk->set_handler(F_AGA, aga_handler);
  }
  
  TSheet_field& sf = _msk->sfield(F_MODULI);
  sf.set_notify(k_notify);

  const int un = _im->unassigned();
  const TDongle::KeyType mydongle = _dongle.type();
  switch (mydongle)
  {
  case TDongle::_user_key:
  {
    _msk->disable(F_K1);
    _msk->disable(F_K2);
    _msk->set_handler(F_K4, user_hnd);
    
    _msk->enable(F_DT);
    _msk->set_handler(F_DT, date_hnd);
    
    generate_key();
//    sf.enable_column(F_ENABLE, FALSE);
    sf.enable_column(F_KEY, serno() != 0);
  }
  break;
  case TDongle::_prassi_key: 
  case TDongle::_aga_key:
  case TDongle::_procom_key:
  {
    _msk->set_handler(F_K2, decode_hnd);
    _msk->disable(F_K3);
    _msk->disable(F_K4); 
    
    _msk->set_handler(F_SN, serno_hnd);
    _msk->enable(F_SN);
    
    sf.sheet_mask().field(F_ENABLE).set_handler(activate_hnd);
    build_sheet(FALSE);  
    
    for (int i = un; i < MAX_AUT; i++)
    {
      const int af = _im->get_module_by_order(i);
      const int index = i-un;
      if (af == 0)
        sf.disable_cell(index,1);            
      else  
        sf.enable_cell(index,1,_im->test_owner(af, mydongle));            
    }
    
    build_key_column();
  }
  break;
  default:
  break;
  }                      
             
  enable_menu_item(M_FILE_PG_SETUP);
  enable_menu_item(M_FILE_PRINT);
  dispatch_e_menu(BAR_ITEM(1));
  return TRUE;
}

bool TAttivazione_moduli::destroy()
{
  if (_msk != NULL) delete _msk;
  if (_im  != NULL) delete _im;
  return TRUE;
}

bool TAttivazione_moduli::menu(MENU_TAG)
{
  KEY res = _msk->run();
  if (res == K_ENTER)
    burn_dongle();

  return FALSE;
}

int ba1500(int argc, char** argv)
{
  // dipende dalla check_parameters fatta in main()
  if (user() == "PRASSI")
  {
    TAttivazione_moduli a ;
    a.run(argc, argv, "Attivazione moduli");
  }
  else 
    error_box("L'utente %s non e' abilitato all'esecuzione di questo programma", (const char*)user());
  return 0;
}
